from compiler import walk
import compiler
import os, os.path
import sys

def _findDottedNamesHelper(node, result):
    more_node = node
    name = node.__class__.__name__
    if name == 'Getattr':
        dotted = []
        while name == 'Getattr':
            dotted.append(node.attrname)
            node = node.expr
            name = node.__class__.__name__
        if name == 'Name':
            dotted.append(node.name)
            dotted.reverse()
            for i in range(1, len(dotted)):
                result.append('.'.join(dotted[:i]))
            result.append('.'.join(dotted))
            return
    elif name == 'Name':
        result.append(node.name)
        return
    elif name == 'AssAttr':
        return 
    for child in more_node.getChildNodes():
        _findDottedNamesHelper(child, result)

def findDottedNames(node):
    result = []
    _findDottedNamesHelper(node, result)
    return result

class ImportFinder:
    def __init__(self):
        self._map = {}
        
    def visitFrom(self, stmt):
        module_name, names = stmt.asList()
        if module_name == '__future__':
            # we don't care what's imported from the future
            return 
        for orig_name, as_name in names:
            # we don't care about from import *
            if orig_name == '*':
                continue
            if as_name is None:
                name = orig_name
            else:
                name = as_name
            self._map.setdefault(module_name, {})[name] = orig_name
    
    def visitImport(self, stmt):
        for orig_name, as_name in stmt.names:
            if as_name is None:
                name = orig_name
            else:
                name = as_name
            self._map.setdefault(orig_name, {})[name] = orig_name

    def getMap(self):
        return self._map
    
def findImports(mod):
    visitor = ImportFinder()
    walk(mod, visitor)
    return visitor.getMap()

class Module:
    def __init__(self, path):
        mod = compiler.parseFile(path)
        self._path = path
        self._map = findImports(mod)
        dottednames = {}
        for name in findDottedNames(mod):
            dottednames[name] = 1
        self._dottednames = dottednames
           
    def getPath(self):
        return self._path
    
    def getModuleNames(self):
        return self._map.keys()

    def getImportInfo(self):
        result = []
        map = self._map
        for module_name in map.keys():
            for usedname, originalname in map[module_name].items():
                result.append((originalname, module_name))
                # add any other name that we could be using
                for dottedname in self._dottednames.keys():
                    usednamedot = usedname + '.'
                    if dottedname.startswith(usednamedot):
                        attrname = dottedname[len(usednamedot):].split('.')[0]
                        result.append((attrname, module_name))
        return result

    def getUnusedNames(self):
        result = []
        for value in self._map.values():
            for usedname, originalname in value.items():
                if usedname not in self._dottednames:
                    result.append(originalname)
        return result
    
class ModuleFinder:
    def __init__(self):
        self._files = []
        
    def visit(self, arg, dirname, names):
        for name in names:
            # get all .py files that aren't weirdo emacs droppings
            if name.endswith('.py') and not name.startswith('.#'):
                self._files.append(os.path.join(dirname, name))

def findModules(path):
    finder = ModuleFinder()
    os.path.walk(path, finder.visit, ())
    return finder._files


class ImportDatabase:
    def __init__(self, root_path):
        self._root_path = root_path
        self._modules = {}
        self._names = {}
        
    def resolveDottedModuleName(self, dotted_name, module):
        """Return path to file representing module, or None if no such thing.
        Can do this relative from module.
        """
        dotted_path = dotted_name.replace('.', '/')
        # try relative import first
        path = os.path.join(os.path.dirname(module.getPath()), dotted_path)
        path = self._resolveHelper(path)
        if path is not None:
            return path
        # absolute import (assumed to be from this tree)
        if os.path.isfile(os.path.join(self._root_path, '__init__.py')):
            startpath, dummy = os.path.split(self._root_path)
        else:
            startpath = self._root_path
        return self._resolveHelper(os.path.join(startpath, dotted_path))

    def _resolveHelper(self, path):
        if os.path.isfile(path + '.py'):
            return path + '.py'
        if os.path.isdir(path):
            path = os.path.join(path, '__init__.py')
            if os.path.isfile(path):
                return path
        return None

    def findModules(self):
        for modulepath in findModules(self._root_path):
            #print modulepath
            module = Module(modulepath)
            self.addModule(module)
        
    def addModule(self, module):
        self_path = module.getPath()
        if self._modules.has_key(self_path):
            return
        self._modules[self_path] = module
        names = self._names
        for name, from_module_name in module.getImportInfo():
            path = self.resolveDottedModuleName(from_module_name, module)
            t = (path, name)
            modulepaths = names.get(t, {})
            if not modulepaths.has_key(self_path):
                modulepaths[self_path] = 1
            names[t] = modulepaths
        
    def getUnusedImports(self):
        """Get all unused imports.
        """
        result = {}
        for path, module in self._modules.items():
            result[path] = self.getUnusedImportsInModule(module)
        return result

    def getUnusedImportsInModule(self, module):
        """Get all unused imports in module.
        """
        result = []
        for name in module.getUnusedNames():
            if not self.isNameImportedFrom(name, module):
                result.append(name)
        return result
    
    def isNameImportedFrom(self, name, module):
        """Return true if name is imported from module by another module.
        """
        return self._names.has_key((module.getPath(), name))
        
    def getModulesImportingNameFrom(self, name, module):
        """Return list of known modules that import name from module.
        """
        result = []
        for path in self._names.get((module.getPath(), name), {}).keys():
            result.append(self._modules[path])
        return result
     
def main():
    try:
        path = sys.argv[1] # '/home/faassen/working/Zope3/src'
    except IndexError:
        print "No path supplied"
        sys.exit(1)
    path = os.path.abspath(path)
    if not os.path.isdir(path):
        print "Unknown path:", path
        sys.exit(1)
    l = len(path) + 1
    db = ImportDatabase(path)
    db.findModules()
    result = []
    for modulepath, names in db.getUnusedImports().items():
        if names:
            result.append((modulepath[l:], names))
    result.sort()
    for modulepath, names in result:
        print modulepath, ' '.join(names)

if __name__ == '__main__':
    main()
    
